/*     CalculiX - A 3-dimensional finite element program                 */
/*              Copyright (C) 1998-2012 Guido Dhondt                          */

/*     This program is free software; you can redistribute it and/or     */
/*     modify it under the terms of the GNU General Public License as    */
/*     published by the Free Software Foundation(version 2);    */
/*                                                                       */

/*     This program is distributed in the hope that it will be useful,   */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of    */ 
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the      */
/*     GNU General Public License for more details.                      */

/*     You should have received a copy of the GNU General Public License */
/*     along with this program; if not, write to the Free Software       */
/*     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.         */

#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include "extUtil.h"
#include "interpol3d.h"

int *nnx1,*nny1,*nnz1,*inodfa1,*ifatet1,nktet1,netet1,*kontet1,
  nfield1,nk1,*kontypm1,*ipkonm1,*konm1,*iparent1,num_cpus;

double *x1,*y_1,*z1,*xo1,*yo1,*zo1,*planfa1,*field1,*cotet1,*co1,*field21;

int main ( int argc, char **argv )
{

  char datin1[MAX_LINE_LENGTH];
  char datin2[MAX_LINE_LENGTH];
  char datout[MAX_LINE_LENGTH];
  char *inpc=NULL,*inpc2=NULL;

  int *kontet=NULL,*ifatet=NULL,*inodfa=NULL,*notet=NULL,
    *ipofa=NULL,type,n1,n2,n3,n4,*nnx=NULL,*nny=NULL,*nnz=NULL,
    *ni=NULL,*kon=NULL,*ipkon=NULL,*kontyp=NULL,*iel=NULL,*no=NULL,
    *iparent=NULL,*kontypm=NULL,*konm=NULL,*ipkonm=NULL,nka,nkb;
  int ifreetet=1,ifreefa=1,iseed=184389,kflag=2,nn,ne,netet,flag,
      indexe,nfield2,nline,nline2;
  int netet_,nktet=0,nfield,j,k,nk,nodes[4],i,ibase,lnode;
  int i1[24]={3,7,8,6,4,3,8,1,3,8,5,6,3,5,8,1,2,3,5,6,2,5,3,1};
  int i2[12]={1,2,3,5,1,5,3,4,4,5,3,6};
  int i4[88]={5,20,17,13,20,8,19,16,19,7,18,15,18,6,17,14,
              1,9,12,13,12,11,4,16,11,10,3,15,9,2,10,14,
              9,13,11,12,11,13,16,12,13,17,19,20,13,19,16,20,
              17,14,19,18,19,14,15,18,9,11,14,10,11,15,14,10,
              11,19,16,13,11,15,19,14,11,14,19,17,11,19,13,17,
              11,14,17,9,11,17,13,9};
  int i5[56]={1,7,9,10,7,2,8,11,8,3,9,12,5,13,14,11,
              13,4,15,10,14,15,6,12,11,12,7,10,7,12,9,10,
              11,7,12,8,12,7,9,8,13,15,11,10,11,15,12,10,
              13,11,15,14,15,11,12,14};
  int i6[32]={8,9,10,4,1,5,7,8,7,6,3,10,9,8,10,7,
              8,9,5,7,9,10,6,7,5,6,7,9,5,2,6,9};

  unsigned int unsiint;

  double *planfa=NULL,*cotet=NULL,*t=NULL,*cgtet=NULL,*field=NULL,*co=NULL,
      xp,yp,zp,*x=NULL,*y=NULL,*z=NULL,*xo=NULL,*yo=NULL,*zo=NULL,*r=NULL,
      *field2=NULL;

  if (argc != 3)
  {
    printf ("\n\n usage: interpol3d master.frd slave.frd \n\n");
    exit (-1);
  }
  else
  {
    strcpy (datin1, argv[1]);
    strcpy (datin2, argv[2]);
    strcpy (datout, "interpol3d.frd");
  }

  flag=1;
  readinput(datin1,&nktet,&ne,&nfield,&konm,&ipkonm,
            &kontypm,&cotet,&field,&iel,&flag,&inpc,&nline,&notet);

  /* generating the tetrahedral elements */
  
  netet=0;
  netet_=22*ne;

  iparent=(int *) u_calloc(netet_,sizeof(int));
  kontet=(int *) u_calloc(4*netet_, sizeof(int));
  ipofa=(int *) u_calloc(4*netet_, sizeof(int));
  inodfa=(int *) u_calloc(16*netet_, sizeof(int));
  ifatet=(int *) u_calloc(4*netet_, sizeof(int));
  planfa=(double *) u_calloc(16*netet_, sizeof(double));

  /* initialization of fields */

  FORTRAN(init,(&nktet,inodfa,ipofa,&netet_));

  for(i=0;i<ne;i++){
    type=kontypm[i];
    indexe=ipkonm[i]-1;
    if(type==1){

      /* C3D8* */

      for(j=0;j<6;j++){
        nodes[0]=konm[indexe+i1[4*j]];
        nodes[1]=konm[indexe+i1[4*j+1]];
        nodes[2]=konm[indexe+i1[4*j+2]];
        nodes[3]=konm[indexe+i1[4*j+3]];
	iparent[netet]=i+1;
        netet++;
        FORTRAN(generatetet,(kontet,ifatet,&netet,inodfa,
                   &ifreefa,planfa,ipofa,nodes,cotet));
      }
    }
    else if(type==2){

      /* C3D6 */

      for(j=0;j<3;j++){
        nodes[0]=konm[indexe+i2[4*j]];
        nodes[1]=konm[indexe+i2[4*j+1]];
        nodes[2]=konm[indexe+i2[4*j+2]];
        nodes[3]=konm[indexe+i2[4*j+3]];
	iparent[netet]=i+1;
        netet++;
        FORTRAN(generatetet,(kontet,ifatet,&netet,inodfa,
                   &ifreefa,planfa,ipofa,nodes,cotet));
      }
    }
    else if(type==3){

      /* C3D4 */

      nodes[0]=konm[indexe+1];
      nodes[1]=konm[indexe+2];
      nodes[2]=konm[indexe+3];
      nodes[3]=konm[indexe+4];
      iparent[netet]=i+1;
      netet++;
      FORTRAN(generatetet,(kontet,ifatet,&netet,inodfa,
               &ifreefa,planfa,ipofa,nodes,cotet));
    }
    else if(type==4){

      /* C3D20* */

      for(j=0;j<22;j++){
        nodes[0]=konm[indexe+i4[4*j]];
        nodes[1]=konm[indexe+i4[4*j+1]];
        nodes[2]=konm[indexe+i4[4*j+2]];
        nodes[3]=konm[indexe+i4[4*j+3]];
	iparent[netet]=i+1;
        netet++;
        FORTRAN(generatetet,(kontet,ifatet,&netet,inodfa,
               &ifreefa,planfa,ipofa,nodes,cotet));
      }
    }
    else if(type==5){

      /* C3D15 */

      for(j=0;j<14;j++){
        nodes[0]=konm[indexe+i5[4*j]];
        nodes[1]=konm[indexe+i5[4*j+1]];
        nodes[2]=konm[indexe+i5[4*j+2]];
        nodes[3]=konm[indexe+i5[4*j+3]];
	iparent[netet]=i+1;
        netet++;
        FORTRAN(generatetet,(kontet,ifatet,&netet,inodfa,
               &ifreefa,planfa,ipofa,nodes,cotet));
      }
   }
    else if(type==6){

      /* C3D10 */

      for(j=0;j<8;j++){
        nodes[0]=konm[indexe+i6[4*j]];
        nodes[1]=konm[indexe+i6[4*j+1]];
        nodes[2]=konm[indexe+i6[4*j+2]];
        nodes[3]=konm[indexe+i6[4*j+3]];
	iparent[netet]=i+1;
        netet++;
        FORTRAN(generatetet,(kontet,ifatet,&netet,inodfa,
               &ifreefa,planfa,ipofa,nodes,cotet));
      }
    }
  }

  /* writing the tet mesh in frd format */
  
  FORTRAN(writetetmesh,(kontet,&netet_,cotet,&nktet,field,&nfield));
              
  /* calculating the center of gravity of the tetrahedra */
              
  cgtet=NNEW(double,3*netet);
  for(i=0;i<netet;i++){
    n1=kontet[4*i]-1;
    n2=kontet[4*i+1]-1;
    n3=kontet[4*i+2]-1;
    n4=kontet[4*i+3]-1;
    cgtet[3*i]=(cotet[3*n1]+cotet[3*n2]+cotet[3*n3]+cotet[3*n4])/4.;
    cgtet[3*i+1]=(cotet[3*n1+1]+cotet[3*n2+1]+cotet[3*n3+1]+cotet[3*n4+1])/4.;
    cgtet[3*i+2]=(cotet[3*n1+2]+cotet[3*n2+2]+cotet[3*n3+2]+cotet[3*n4+2])/4.;
  }

  free(iel);free(no);

  /* reading the nodes in which the fields are to be interpolated,
     and related information */

  flag=0;
  readinput(datin2,&nk,&ne,&nfield2,&kon,&ipkon,
            &kontyp,&co,&field2,&iel,&flag,&inpc2,&nline2,&no);

  /* initialization of additional fields */
  
  x=NNEW(double,netet);
  y=NNEW(double,netet);
  z=NNEW(double,netet);
  xo=NNEW(double,netet);
  yo=NNEW(double,netet);
  zo=NNEW(double,netet);
  r=NNEW(double,netet);
  nnx=NNEW(int,netet);
  nny=NNEW(int,netet);
  nnz=NNEW(int,netet);
  ni=NNEW(int,netet);
  for(i=0;i<netet;i++){
    nnx[i]=i+1;
    nny[i]=i+1;
    nnz[i]=i+1;
    x[i]=cgtet[3*i];
    y[i]=cgtet[3*i+1];
    z[i]=cgtet[3*i+2];
    xo[i]=x[i];
    yo[i]=y[i];
    zo[i]=z[i];
  }
  FORTRAN(dsort,(x,nnx,&netet,&kflag));
  FORTRAN(dsort,(y,nny,&netet,&kflag));
  FORTRAN(dsort,(z,nnz,&netet,&kflag));

  /* loop over all nodes in which the temperatures have to be
     interpolated */

  t=NNEW(double,nfield);
  RENEW(field2,double,nfield*nk);

  /* variables for multithreading procedure */

  int sys_cpus;
  char *env,*envloc,*envsys;
      
  num_cpus = 0;
  sys_cpus=0;
  
  /* explicit user declaration prevails */
  
  envsys=getenv("OMP_NUM_THREADS");
  if(envsys){
      sys_cpus=atoi(envsys);
      if(sys_cpus<0) sys_cpus=0;
  }
    
  /* local declaration prevails, if strictly positive */
  
  envloc = getenv("OMP_NUM_THREADS");
  if(envloc){
      num_cpus=atoi(envloc);
      if(num_cpus<1){
	  num_cpus=1;
      }else if(num_cpus>sys_cpus){
	  num_cpus=sys_cpus;
      }
  }else{
    num_cpus=sys_cpus;
  }
    
  if(nk<num_cpus) num_cpus=nk;
  
  printf(" Using up to %d cpu(s) for interpol3d.\n", num_cpus);
  
  pthread_t tid[num_cpus];
  
  x1=x;y_1=y;z1=z;xo1=xo;yo1=yo;zo1=zo;nnx1=nnx;nny1=nny;nnz1=nnz;
  inodfa1=inodfa;planfa1=planfa;ifatet1=ifatet;nktet1=nktet;netet1=netet;
  kontet1=kontet;field1=field;nfield1=nfield;cotet1=cotet;nk1=nk;
  kontypm1=kontypm;ipkonm1=ipkonm;konm1=konm;iparent1=iparent;co1=co;
  field21=field2;
  
  /* create threads and wait */
  
  for(i=0; i<num_cpus; i++)  {
    pthread_create(&tid[i], NULL, basist_3dmt, (void *)i);
  }
  for(i=0; i<num_cpus; i++)  pthread_join(tid[i], NULL);

  /* storing the interpolated values */


  FORTRAN(writemesh,(kon,&ne,co,&nk,field2,&nfield,inpc,&nline,
                     no,iel,ipkon,kontyp));

  return(0);

}

/* subroutine for multithreading of basist_3d */

void *basist_3dmt(void *i){

  int nka,nkb,nkdelta;
    
  nkdelta=(int)ceil(nk1/(double)num_cpus);
  nka=((int)i)*nkdelta+1;
  nkb=(((int)i)+1)*nkdelta;
  if(nkb>nk1) nkb=nk1;

  FORTRAN(basist_3d,(x1,y_1,z1,xo1,yo1,zo1,nnx1,nny1,nnz1,
              inodfa1,planfa1,ifatet1,&nktet1,&netet1,kontet1,field1,&nfield1,
	      cotet1,&nk1,kontypm1,ipkonm1,konm1,iparent1,co1,field21,
              &nka,&nkb));

    return NULL;
}
